@@ -3,12 +3,13 @@
 from sql import Null
 
 from trytond.model import Workflow, Model, ModelView, ModelSQL, fields, Check
-from trytond.pyson import Eval, Bool
-from trytond import backend
+from trytond.pyson import Eval, Bool, If
 from trytond.transaction import Transaction
 from trytond.pool import Pool
+from trytond.wizard import Wizard, StateView, StateTransition, Button
 
-__all__ = ['Inventory', 'InventoryLine']
+__all__ = ['Inventory', 'InventoryLine',
+    'Count', 'CountSearch', 'CountQuantity']
 
 STATES = {
     'readonly': Eval('state') != 'draft',
@@ -40,8 +41,18 @@ class Inventory(Workflow, ModelSQL, ModelView):
         'stock.location', 'Lost and Found', required=True,
         domain=[('type', '=', 'lost_found')], states=STATES, depends=DEPENDS)
     lines = fields.One2Many(
-        'stock.inventory.line', 'inventory', 'Lines', states=STATES,
-        depends=DEPENDS)
+        'stock.inventory.line', 'inventory', 'Lines',
+        states={
+            'readonly': (STATES['readonly'] | ~Eval('location')
+                | ~Eval('date')),
+            },
+        depends=['location', 'date'] + DEPENDS)
+    empty_quantity = fields.Selection([
+            (None, ""),
+            ('keep', "Keep"),
+            ('empty', "Empty"),
+            ], "Empty Quantity", states=STATES, depends=DEPENDS,
+        help="How lines without quantity are handled.")
     company = fields.Many2One('company.company', 'Company', required=True,
         states={
             'readonly': (Eval('state') != 'draft') | Eval('lines', [0]),
@@ -55,7 +66,7 @@ class Inventory(Workflow, ModelSQL, ModelView):
         super(Inventory, cls).__setup__()
         cls._order.insert(0, ('date', 'DESC'))
         cls._error_messages.update({
-                'delete_cancel': ('Inventory "%s" must be cancelled before '
+                'delete_cancel': ('Inventory "%s" must be canceled before '
                     'deletion.'),
                 'unique_line': ('Line "%s" is not unique '
                     'on Inventory "%s".'),
@@ -67,22 +78,29 @@ class Inventory(Workflow, ModelSQL, ModelView):
         cls._buttons.update({
                 'confirm': {
                     'invisible': Eval('state').in_(['done', 'cancel']),
+                    'depends': ['state'],
                     },
                 'cancel': {
                     'invisible': Eval('state').in_(['cancel', 'done']),
+                    'depends': ['state'],
                     },
                 'complete_lines': {
                     'readonly': Eval('state') != 'draft',
+                    'depends': ['state'],
+                    },
+                'count': {
+                    'readonly': Eval('state') != 'draft',
+                    'depends': ['state'],
                     },
                 })
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
         super(Inventory, cls).__register__(module_name)
 
+        table = cls.__table_handler__(module_name)
+
         # Add index on create_date
-        table = TableHandler(cls, module_name)
         table.index_action('create_date', action='add')
 
     @staticmethod
@@ -166,26 +184,17 @@ class Inventory(Workflow, ModelSQL, ModelView):
     def copy(cls, inventories, default=None):
         pool = Pool()
         Date = pool.get('ir.date')
-        Line = pool.get('stock.inventory.line')
 
         if default is None:
             default = {}
-        default = default.copy()
-        default['date'] = Date.today()
-        default['lines'] = None
-        default.setdefault('number')
+        else:
+            default = default.copy()
+        default.setdefault('date', Date.today())
+        default.setdefault('lines.moves', None)
+        default.setdefault('number', None)
 
-        new_inventories = []
-        for inventory in inventories:
-            new_inventory, = super(Inventory, cls).copy([inventory],
-                default=default)
-            Line.copy(inventory.lines,
-                default={
-                    'inventory': new_inventory.id,
-                    'moves': None,
-                    })
-            cls.complete_lines([new_inventory], fill=False)
-            new_inventories.append(new_inventory)
+        new_inventories = super().copy(inventories, default=default)
+        cls.complete_lines(new_inventories, fill=False)
         return new_inventories
 
     @staticmethod
@@ -215,8 +224,9 @@ class Inventory(Workflow, ModelSQL, ModelView):
                 product_ids = [l.product.id for l in inventory.lines]
             with Transaction().set_context(stock_date_end=inventory.date):
                 pbl = Product.products_by_location(
-                    [inventory.location.id], product_ids=product_ids,
-                    grouping=grouping)
+                    [inventory.location.id],
+                    grouping=grouping,
+                    grouping_filter=(product_ids,))
 
             # Index some data
             product2type = {}
@@ -244,7 +254,7 @@ class Inventory(Workflow, ModelSQL, ModelView):
             if not fill:
                 continue
             # Create lines if needed
-            for key, quantity in pbl.iteritems():
+            for key, quantity in pbl.items():
                 product_id = key[grouping.index('product') + 1]
                 if (product2type[product_id] != 'goods'
                         or product2consumable[product_id]):
@@ -261,6 +271,11 @@ class Inventory(Workflow, ModelSQL, ModelView):
         if to_write:
             Line.write(*to_write)
 
+    @classmethod
+    @ModelView.button_action('stock.wizard_inventory_count')
+    def count(cls, inventories):
+        cls.complete_lines(inventories)
+
 
 class InventoryLine(ModelSQL, ModelView):
     'Stock Inventory Line'
@@ -279,9 +294,12 @@ class InventoryLine(ModelSQL, ModelView):
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
             'get_unit_digits')
     expected_quantity = fields.Float('Expected Quantity', required=True,
-            digits=(16, Eval('unit_digits', 2)), readonly=True,
-            depends=['unit_digits'])
-    quantity = fields.Float('Quantity', required=True,
+        digits=(16, Eval('unit_digits', 2)), readonly=True,
+        states={
+            'invisible': Eval('id', -1) < 0,
+        },
+        depends=['unit_digits'])
+    quantity = fields.Float('Quantity',
         digits=(16, Eval('unit_digits', 2)),
         states=_states, depends=['unit_digits'] + _depends)
     moves = fields.One2Many('stock.move', 'origin', 'Moves', readonly=True)
@@ -304,10 +322,15 @@ class InventoryLine(ModelSQL, ModelView):
                 'Line quantity must be positive.'),
             ]
         cls._order.insert(0, ('product', 'ASC'))
+        cls._error_messages.update({
+                'missing_empty_quantity': ('An option for empty quantity is '
+                    'missing for inventory "%(inventory)s".'),
+                'delete_cancel_draft': ('The line "%(line)s" must be on '
+                    'canceled or draft inventory to be deleted.'),
+                })
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
         cursor = Transaction().connection.cursor()
         pool = Pool()
         Move = pool.get('stock.move')
@@ -316,9 +339,7 @@ class InventoryLine(ModelSQL, ModelView):
 
         super(InventoryLine, cls).__register__(module_name)
 
-        table = TableHandler(cls, module_name)
-        # Migration from 2.8: Remove constraint inventory_product_uniq
-        table.drop_constraint('inventory_product_uniq')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 3.0: use Move origin
         if table.column_exist('move'):
@@ -331,6 +352,9 @@ class InventoryLine(ModelSQL, ModelView):
                         where=move_table.id == move_id))
             table.drop_column('move')
 
+        # Migration from 4.6: drop required on quantity
+        table.not_null_action('quantity', action='remove')
+
     @staticmethod
     def default_unit_digits():
         return 2
@@ -390,8 +414,19 @@ class InventoryLine(ModelSQL, ModelView):
         Move = pool.get('stock.move')
         Uom = pool.get('product.uom')
 
+        qty = self.quantity
+        if qty is None:
+            if self.inventory.empty_quantity is None:
+                self.raise_user_error('missing_empty_quantity', {
+                        'inventory': self.inventory.rec_name,
+                        })
+            if self.inventory.empty_quantity == 'keep':
+                return
+            else:
+                qty = 0.0
+
         delta_qty = Uom.compute_qty(self.uom,
-            self.expected_quantity - self.quantity,
+            self.expected_quantity - qty,
             self.uom)
         if delta_qty == 0.0:
             return
@@ -421,9 +456,6 @@ class InventoryLine(ModelSQL, ModelView):
         if self.quantity == self.expected_quantity == quantity:
             return values
         values['expected_quantity'] = quantity
-        # update also quantity field if not edited
-        if self.quantity == self.expected_quantity:
-            values['quantity'] = max(quantity, 0.0)
         return values
 
     @classmethod
@@ -434,5 +466,153 @@ class InventoryLine(ModelSQL, ModelView):
         return {
             'inventory': inventory.id,
             'expected_quantity': quantity,
-            'quantity': max(quantity, 0.0),
         }
+
+    @classmethod
+    def delete(cls, lines):
+        for line in lines:
+            if line.inventory_state not in {'cancel', 'draft'}:
+                cls.raise_user_error('delete_cancel_draft', {
+                        'line': line.rec_name,
+                        })
+        super(InventoryLine, cls).delete(lines)
+
+
+class Count(Wizard):
+    "Stock Inventory Count"
+    __name__ = 'stock.inventory.count'
+    start_state = 'search'
+
+    search = StateView(
+        'stock.inventory.count.search',
+        'stock.inventory_count_search_view_form', [
+            Button("End", 'end', 'tryton-cancel'),
+            Button("Select", 'quantity', 'tryton-forward', default=True),
+            ])
+    quantity = StateView(
+        'stock.inventory.count.quantity',
+        'stock.inventory_count_quantity_view_form', [
+            Button("Cancel", 'search', 'tryton-cancel'),
+            Button("Add", 'add', 'tryton-ok', default=True),
+            ])
+    add = StateTransition()
+
+    @classmethod
+    def __setup__(cls):
+        super(Count, cls).__setup__()
+        cls._error_messages.update({
+                'create_line': "No existing line found for %(search)s.",
+                })
+
+    def default_quantity(self, fields):
+        pool = Pool()
+        Inventory = pool.get('stock.inventory')
+        InventoryLine = pool.get('stock.inventory.line')
+        context = Transaction().context
+        inventory = Inventory(context['active_id'])
+        values = {}
+        lines = InventoryLine.search(self.get_line_domain(inventory), limit=1)
+        if not lines:
+            warning_name = '%s.%s.count_create' % (
+                inventory, self.search.search)
+            self.raise_user_warning(warning_name, 'create_line', {
+                    'search': self.search.search.rec_name,
+                    })
+            line, = InventoryLine.create([self.get_line_values(inventory)])
+        else:
+            line, = lines
+        values['line'] = line.id
+        values['product'] = line.product.id
+        values['uom'] = line.uom.id
+        values['unit_digits'] = line.unit_digits
+        if line.uom.rounding == 1:
+            values['quantity_added'] = 1
+        return values
+
+    def get_line_domain(self, inventory):
+        pool = Pool()
+        Product = pool.get('product.product')
+        domain = [
+            ('inventory', '=', inventory.id),
+            ]
+        if isinstance(self.search.search, Product):
+            domain.append(('product', '=', self.search.search.id))
+        return domain
+
+    def get_line_values(self, inventory):
+        pool = Pool()
+        Product = pool.get('product.product')
+        InventoryLine = pool.get('stock.inventory.line')
+        values = InventoryLine.create_values4complete(inventory, 0)
+        if isinstance(self.search.search, Product):
+            values['product'] = self.search.search.id
+        return values
+
+    def transition_add(self):
+        if self.quantity.line and self.quantity.quantity_added:
+            line = self.quantity.line
+            if line.quantity:
+                line.quantity += self.quantity.quantity_added
+            else:
+                line.quantity = self.quantity.quantity_added
+            line.save()
+        return 'search'
+
+
+class CountSearch(ModelView):
+    "Stock Inventory Count"
+    __name__ = 'stock.inventory.count.search'
+
+    search = fields.Reference(
+        "Search", [
+            ('product.product', "Product"),
+            ],
+        required=True,
+        domain=[If(Eval('search_model') == 'product.product',
+                [
+                    ('type', '=', 'goods'),
+                    ('consumable', '=', False),
+                    ],
+                [])],
+        depends=['search_model'])
+    search_model = fields.Function(fields.Selection(
+        'get_search_models', "Search Model"),
+        'on_change_with_search_model')
+
+    @classmethod
+    def default_search(cls):
+        return 'product.product,-1'
+
+    @classmethod
+    def get_search_models(cls):
+        return cls.fields_get(['search'])['search']['selection']
+
+    @fields.depends('search')
+    def on_change_with_search_model(self, name=None):
+        if self.search:
+            return self.search.__name__
+
+
+class CountQuantity(ModelView):
+    "Stock Inventory Count"
+    __name__ = 'stock.inventory.count.quantity'
+
+    line = fields.Many2One(
+        'stock.inventory.line', "Line", readonly=True, required=True)
+    product = fields.Many2One('product.product', "Product", readonly=True)
+    uom = fields.Many2One('product.uom', "UOM", readonly=True)
+    quantity_resulting = fields.Float(
+        "Resulting Quantity", digits=(16, Eval('unit_digits', 2)),
+        readonly=True, depends=['unit_digits'])
+
+    quantity_added = fields.Float(
+        "Added Quantity", digits=(16, Eval('unit_digits', 2)), required=True,
+        depends=['unit_digits'])
+
+    unit_digits = fields.Integer("Unit Digits", readonly=True)
+
+    @fields.depends('quantity_added', 'line')
+    def on_change_quantity_added(self):
+        if self.line:
+            self.quantity_resulting = (
+                (self.line.quantity or 0) + (self.quantity_added or 0))
