@@ -1,6 +1,7 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 import datetime
+import functools
 from decimal import Decimal
 from collections import defaultdict
 
@@ -25,8 +26,44 @@ __all__ = ['Template', 'Product',
     'RecomputeCostPrice']
 
 
-class Template:
-    __metaclass__ = PoolMeta
+def check_no_move(func):
+
+    def find_moves(cls, records):
+        pool = Pool()
+        Move = pool.get('stock.move')
+        if cls.__name__ == 'product.template':
+            field = 'product.template'
+        else:
+            field = 'product'
+        for sub_records in grouped_slice(records):
+            moves = Move.search([
+                    (field, 'in', list(map(int, sub_records))),
+                    ],
+                limit=1, order=[])
+            if moves:
+                return True
+        return False
+
+    @functools.wraps(func)
+    def decorator(cls, *args):
+        pool = Pool()
+        Template = pool.get('product.template')
+        transaction = Transaction()
+        if (transaction.user != 0
+                and transaction.context.get('_check_access')):
+            actions = iter(args)
+            for records, values in zip(actions, actions):
+                for field, error in Template._modify_no_move:
+                    if field in values:
+                        if find_moves(cls, records):
+                            Template.raise_user_error(error)
+                        # No moves for those records
+                        break
+        func(cls, *args)
+    return decorator
+
+
+class Template(metaclass=PoolMeta):
     __name__ = "product.template"
     quantity = fields.Function(fields.Float('Quantity'), 'sum_product')
     forecast_quantity = fields.Function(fields.Float('Forecast Quantity'),
@@ -61,26 +98,8 @@ class Template:
             ]
 
     @classmethod
-    def check_no_move(cls, templates, error):
-        Move = Pool().get('stock.move')
-        for sub_templates in grouped_slice(templates):
-            moves = Move.search([
-                    ('product.template', 'in', [t.id for t in sub_templates]),
-                    ],
-                limit=1, order=[])
-            if moves:
-                cls.raise_user_error(error)
-
-    @classmethod
+    @check_no_move
     def write(cls, *args):
-        if (Transaction().user != 0
-                and Transaction().context.get('_check_access')):
-            actions = iter(args)
-            for templates, values in zip(actions, actions):
-                for field, error in cls._modify_no_move:
-                    if field in values:
-                        cls.check_no_move(templates, error)
-                        break
         super(Template, cls).write(*args)
 
     @classmethod
@@ -92,8 +111,7 @@ class Template:
         Product.recompute_cost_price(products)
 
 
-class Product(StockMixin, object):
-    __metaclass__ = PoolMeta
+class Product(StockMixin, object, metaclass=PoolMeta):
     __name__ = "product.product"
     quantity = fields.Function(fields.Float('Quantity'), 'get_quantity',
             searcher='search_quantity')
@@ -105,7 +123,9 @@ class Product(StockMixin, object):
     @classmethod
     def get_quantity(cls, products, name):
         location_ids = Transaction().context.get('locations')
-        return cls._get_quantity(products, name, location_ids, products)
+        product_ids = list(map(int, products))
+        return cls._get_quantity(
+            products, name, location_ids, grouping_filter=(product_ids,))
 
     @classmethod
     def search_quantity(cls, name, domain=None):
@@ -114,26 +134,32 @@ class Product(StockMixin, object):
 
     @classmethod
     def get_cost_value(cls, products, name):
-        cost_values = {}
+        cost_values = {p.id: None for p in products}
         context = {}
         trans_context = Transaction().context
         if trans_context.get('stock_date_end'):
             # Use the last cost_price of the day
             context['_datetime'] = datetime.datetime.combine(
                 trans_context['stock_date_end'], datetime.time.max)
+            # The date could be before the product creation
+            products = [p for p in products
+                if p.create_date <= context['_datetime']]
         with Transaction().set_context(context):
             for product in cls.browse(products):
-                # The date could be before the product creation
-                if not isinstance(product.cost_price, Decimal):
-                    cost_values[product.id] = None
-                else:
-                    cost_values[product.id] = (Decimal(str(product.quantity))
-                        * product.cost_price)
+                # The product may not have a cost price
+                if product.cost_price is not None:
+                    cost_values[product.id] = (
+                        Decimal(str(product.quantity)) * product.cost_price)
         return cost_values
 
     @classmethod
-    def products_by_location(cls, location_ids, product_ids=None,
-            with_childs=False, grouping=('product',)):
+    @check_no_move
+    def write(cls, *args):
+        super(Product, cls).write(*args)
+
+    @classmethod
+    def products_by_location(cls, location_ids,
+            with_childs=False, grouping=('product',), grouping_filter=None):
         """
         Compute for each location and product the stock quantity in the default
         uom of the product.
@@ -166,7 +192,6 @@ class Product(StockMixin, object):
                     wh_to_add[location.id] = location.storage_location.id
             location_ids = list(location_ids)
 
-        grouping_filter = (product_ids,) + tuple(None for k in grouping[1:])
         query = Move.compute_quantities_query(location_ids, with_childs,
             grouping=grouping, grouping_filter=grouping_filter)
         if query is None:
@@ -175,7 +200,7 @@ class Product(StockMixin, object):
             grouping=grouping, grouping_filter=grouping_filter)
 
         if wh_to_add:
-            for wh, storage in wh_to_add.iteritems():
+            for wh, storage in wh_to_add.items():
                 for key in quantities:
                     if key[0] == storage:
                         quantities[(wh,) + key[1:]] = quantities[key]
@@ -199,7 +224,7 @@ class Product(StockMixin, object):
             return
 
         to_write = []
-        for cost, records in costs.iteritems():
+        for cost, records in costs.items():
             to_write.append(records)
             to_write.append({'cost_price': cost})
 
@@ -303,23 +328,32 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
         pool = Pool()
         Move = pool.get('stock.move')
         Location = pool.get('stock.location')
-        move = Move.__table__()
+        Product = pool.get('product.product')
+        move = from_ = Move.__table__()
+        context = Transaction().context
 
-        product_id = Transaction().context.get('product')
-        warehouse_id = Transaction().context.get('warehouse', -1)
+        if context.get('product_template') is not None:
+            product = Product.__table__()
+            from_ = move.join(product, condition=move.product == product.id)
+            product_clause = (
+                product.template == context['product_template'])
+        else:
+            product_clause = move.product == context.get('product', -1)
+
+        warehouse_id = context.get('warehouse', -1)
         warehouse_query = Location.search([
                 ('parent', 'child_of', [warehouse_id]),
                 ], query=True, order=[])
         date_column = Coalesce(move.effective_date, move.planned_date
             ).as_('date')
-        return move.select(
+        return from_.select(
             Max(move.id).as_('id'),
             Literal(0).as_('create_uid'),
             CurrentTimestamp().as_('create_date'),
             Literal(None).as_('write_uid'),
             Literal(None).as_('write_date'),
             date_column,
-            where=(move.product == product_id)
+            where=product_clause
             & (move.from_location.in_(warehouse_query)
                 | move.to_location.in_(warehouse_query))
             & (Coalesce(move.effective_date, move.planned_date) != Null),
@@ -328,9 +362,24 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
     @classmethod
     def get_quantity(cls, lines, name):
         Product = Pool().get('product.product')
+        trans_context = Transaction().context
 
-        product_id = Transaction().context.get('product')
-        warehouse_id = Transaction().context.get('warehouse')
+        def valid_context(name):
+            return (trans_context.get(name) is not None
+                and isinstance(trans_context[name], int))
+
+        if not any(map(valid_context, ['product', 'product_template'])):
+            return {l.id: None for l in lines}
+
+        if trans_context.get('product') is not None:
+            grouping = ('product',)
+            grouping_filter = ([trans_context['product']],)
+            key = trans_context['product']
+        else:
+            grouping = ('product.template',)
+            grouping_filter = ([trans_context.get('product_template')],)
+            key = trans_context['product_template']
+        warehouse_id = trans_context.get('warehouse')
 
         dates = sorted(l.date for l in lines)
         quantities = {}
@@ -343,8 +392,10 @@ class ProductQuantitiesByWarehouse(ModelSQL, ModelView):
                 }
             with Transaction().set_context(**context):
                 quantities[date] = Product.products_by_location(
-                    [warehouse_id], [product_id],
-                    with_childs=True).get((warehouse_id, product_id), 0)
+                    [warehouse_id],
+                    grouping=grouping,
+                    grouping_filter=grouping_filter,
+                    with_childs=True).get((warehouse_id, key), 0)
             try:
                 date_start = date + datetime.timedelta(1)
             except OverflowError:
